# -*- coding: utf-8 -*-
"""
test_compound
~~~~~~~~~~~~~

Test the Compound object.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import logging
import os

import pytest
import requests

from chemspipy import ChemSpider, Compound


logging.basicConfig(level=logging.WARN, format='%(levelname)s:%(name)s:(%(threadName)-10s):%(message)s')
logging.getLogger('chemspipy').setLevel(logging.DEBUG)

# API key is retrieved from environment variables
CHEMSPIDER_API_KEY = os.environ['CHEMSPIDER_API_KEY']
cs = ChemSpider(CHEMSPIDER_API_KEY)


def test_get_compound():
    """Test getting a compound by ChemSpider ID."""
    compound = cs.get_compound(2157)
    assert isinstance(compound, Compound)
    assert compound.record_id == 2157
    with pytest.deprecated_call():
        assert compound.csid == 2157
    compound = cs.get_compound('2157')
    assert isinstance(compound, Compound)
    assert compound.record_id == 2157
    with pytest.deprecated_call():
        assert compound.csid == 2157


def test_get_compounds():
    """Test getting multiple compounds by ChemSpider ID."""
    compounds = cs.get_compounds([2157, 13837760])
    assert [c.csid for c in compounds], [2157 == 13837760]
    for c in compounds:
        assert 'http://' in c.image_url
        assert c.average_mass > 0


def test_compound_init():
    """Test instantiating a Compound directly."""
    compound = Compound(cs, 2157)
    assert compound.csid == 2157


def test_compound_equality():
    """Test equality test by ChemSpider ID."""
    c1 = cs.get_compound(13837760)
    c2 = cs.get_compound(2157)
    c3 = cs.get_compound(2157)
    assert c1 != c2
    assert c2 == c3


def test_compound_repr():
    """Test Compound object repr."""
    assert repr(cs.get_compound(1234)) == 'Compound(1234)'


def test_image_url():
    """Test image_url returns a valid URL."""
    url = cs.get_compound(2157).image_url
    response = requests.get(url)
    assert 'http://www.chemspider.com/ImagesHandler.ashx?id=' in url
    assert response.status_code == 200


def test_molecular_formula():
    """Test Compound property molecular_formula."""
    compound = cs.get_compound(2157)
    assert compound.molecular_formula == 'C_{9}H_{8}O_{4}'
    # Ensure value is the same on subsequent access from cache
    assert compound.molecular_formula == 'C_{9}H_{8}O_{4}'


def test_smiles():
    """Test Compound property smiles."""
    compound = cs.get_compound(2157)
    assert compound.smiles == 'CC(=O)Oc1ccccc1C(=O)O'
    # Ensure value is the same on subsequent access from cache
    assert compound.smiles == 'CC(=O)Oc1ccccc1C(=O)O'


def test_inchi():
    """Test Compound property inchi."""
    compound = cs.get_compound(2157)
    assert compound.inchi == 'InChI=1S/C9H8O4/c1-6(10)13-8-5-3-2-4-7(8)9(11)12/h2-5H,1H3,(H,11,12)'
    # Ensure value is the same on subsequent access from cache
    assert compound.inchi == 'InChI=1S/C9H8O4/c1-6(10)13-8-5-3-2-4-7(8)9(11)12/h2-5H,1H3,(H,11,12)'


def test_stdinchi():
    """Test Compound property stdinchi."""
    compound = cs.get_compound(2157)
    with pytest.deprecated_call():
        assert compound.stdinchi == 'InChI=1S/C9H8O4/c1-6(10)13-8-5-3-2-4-7(8)9(11)12/h2-5H,1H3,(H,11,12)'
        # Ensure value is the same on subsequent access from cache
        assert compound.stdinchi == 'InChI=1S/C9H8O4/c1-6(10)13-8-5-3-2-4-7(8)9(11)12/h2-5H,1H3,(H,11,12)'


def test_inchikey():
    """Test Compound property inchikey."""
    compound = cs.get_compound(2157)
    assert compound.inchikey == 'BSYNRYMUTXBXSQ-UHFFFAOYSA-N'
    # Ensure value is the same on subsequent access from cache
    assert compound.inchikey == 'BSYNRYMUTXBXSQ-UHFFFAOYSA-N'


def test_stdinchikey():
    """Test Compound property stdinchikey."""
    compound = cs.get_compound(2157)
    with pytest.deprecated_call():
        assert compound.stdinchikey == 'BSYNRYMUTXBXSQ-UHFFFAOYSA-N'
        # Ensure value is the same on subsequent access from cache
        assert compound.stdinchikey == 'BSYNRYMUTXBXSQ-UHFFFAOYSA-N'


def test_masses():
    """Test Compound property average_mass, molecular_weight, monoisotopic_mass, nominal_mass."""
    compound = cs.get_compound(2157)
    assert 180 < compound.average_mass < 180.2
    assert 180 < compound.molecular_weight < 180.2
    assert 180 < compound.monoisotopic_mass < 180.2
    assert compound.nominal_mass == 180


def test_name():
    """Test Compound property common_name."""
    compound = cs.get_compound(2157)
    assert compound.common_name == 'Aspirin'


def test_molfiles():
    """Test Compound property mol2d, mol3d, mol_raw."""
    compound = cs.get_compound(2157)
    assert 'V2000' in compound.mol_2d
    assert 'V2000' in compound.mol_3d


def test_image():
    """Test Compound property image."""
    compound = cs.get_compound(2157)
    assert compound.image[:8] == b'\x89PNG\x0d\x0a\x1a\x0a'  # PNG magic number


def test_external_references():
    """Test Compound property external_references."""
    compound = cs.get_compound(97809)
    assert len(compound.external_references) > 50
    for xref in compound.external_references:
        assert 'externalId' in xref
        assert 'externalUrl' in xref
        assert 'source' in xref
        assert 'sourceUrl' in xref



# -*- coding: utf-8 -*-
"""
test_utils
~~~~~~~~~~

Test miscellaneous utility functions.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import datetime
import logging

from chemspipy.utils import timestamp, duration


logging.basicConfig(level=logging.WARN, format='%(levelname)s:%(name)s:(%(threadName)-10s):%(message)s')
logging.getLogger('chemspipy').setLevel(logging.DEBUG)


def test_timestamp_microseconds():
    """Test timestamp parser function on timestamps strings with microseconds."""
    assert timestamp('2007-08-08T20:18:36.593') == datetime.datetime(2007, 8, 8, 20, 18, 36, 593000)
    assert timestamp('2035-12-31T23:59:59.999') == datetime.datetime(2035, 12, 31, 23, 59, 59, 999000)


def test_timestamp_seconds():
    """Test timestamp parser function on timestamps strings with no microseconds."""
    assert timestamp('2007-08-08T20:18:36') == datetime.datetime(2007, 8, 8, 20, 18, 36)
    assert timestamp('2010-09-01T04:33:59') == datetime.datetime(2010, 9, 1, 4, 33, 59)


def test_duration_microseconds():
    """Test duration parser function on duration strings with microseconds."""
    assert duration('0:00:00.001') == datetime.timedelta(0, 0, 1000)
    assert duration('0:00:00.12') == datetime.timedelta(0, 0, 120000)
    assert duration('0:00:00.120') == datetime.timedelta(0, 0, 120000)
    assert duration('0:00:00.052') == datetime.timedelta(0, 0, 52000)
    assert duration('0:00:03.523') == datetime.timedelta(0, 3, 523000)


def test_duration_seconds():
    """Test duration parser function on duration strings with no microseconds."""
    assert duration('0:00:00') == datetime.timedelta(0)
    assert duration('0:00:03') == datetime.timedelta(0, 3)



# -*- coding: utf-8 -*-
"""
test_api
~~~~~~~~

Test the core API functionality.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import logging
import os
import re
import time

import pytest
import six

from chemspipy import ChemSpider, errors


logging.basicConfig(level=logging.WARN)
logging.getLogger('chemspipy').setLevel(logging.DEBUG)

# API key is retrieved from environment variables
CHEMSPIDER_API_KEY = os.environ['CHEMSPIDER_API_KEY']

# Chemspider instances with and without an API key
cs = ChemSpider(CHEMSPIDER_API_KEY)


def test_no_api_key():
    """Test ChemSpider cannot be initialized with no API key."""
    with pytest.raises(TypeError):
        ChemSpider()


def test_api_key():
    """Test API key is set correctly when initializing ChemSpider."""
    assert cs.api_key == CHEMSPIDER_API_KEY


def test_chemspider_repr():
    """Test ChemSpider object repr."""
    assert repr(cs) == 'ChemSpider()'


# Lookups

def test_get_datasources():
    """Test get_datasources returns the list of ChemSpider data sources."""
    datasources = cs.get_datasources()
    assert all(source in datasources for source in ['Wikipedia', 'ZINC', 'PubChem'])


# Records

def test_get_details():
    """Test get_details returns details for a record ID."""
    info = cs.get_details(6543)
    assert all(field in info for field in [
        'id', 'smiles', 'formula', 'averageMass', 'molecularWeight', 'monoisotopicMass', 'nominalMass',
        'commonName', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount', 'mol2D'
    ])
    assert all(isinstance(info[field], float) for field in [
        'averageMass', 'molecularWeight', 'monoisotopicMass'
    ])
    assert isinstance(info['id'], int)
    assert all(isinstance(info[field], six.text_type) for field in [
        'smiles', 'formula', 'commonName', 'mol2D'
    ])


def test_get_details_batch():
    """Test get_extended_compound_info_list returns info for a list of record IDs."""
    info = cs.get_details_batch([6543, 1235, 6084])
    assert len(info) == 3
    assert all(field in info[0] for field in [
        'id', 'smiles', 'formula', 'averageMass', 'molecularWeight', 'monoisotopicMass', 'nominalMass',
        'commonName', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount', 'mol2D'
    ])
    assert all(isinstance(info[0][field], float) for field in [
        'averageMass', 'molecularWeight', 'monoisotopicMass'
    ])
    assert isinstance(info[0]['id'], int)
    assert all(isinstance(info[0][field], six.text_type) for field in [
        'smiles', 'formula', 'commonName', 'mol2D'
    ])


def test_get_external_references():
    """Test get_external_references returns references for a record ID."""
    refs = cs.get_external_references(125)
    assert len(refs) > 5
    for ref in refs:
        assert 'source' in ref
        assert 'sourceUrl' in ref
        assert 'externalId' in ref
        assert 'externalUrl' in ref


def test_get_image():
    """Test get_image returns image data for a record ID."""
    img = cs.get_image(123)
    assert img[:8] == b'\x89PNG\x0d\x0a\x1a\x0a'  # PNG magic number


def test_get_mol():
    """Test get_mol returns a MOLfile for a record ID."""
    mol = cs.get_mol(6084)
    assert 'V2000' in mol
    assert 'M  END' in mol


# Filter

def test_filter_formula_batch():
    """Test filter_formula_batch returns a list of CSIDs."""
    qid = cs.filter_formula_batch(formulas=['C2H2', 'C3H6'])
    while True:
        status = cs.filter_formula_batch_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_formula_batch_results(qid)
    assert len(results) == 2
    for result in results:
        assert 'formula' in result
        assert 'results' in result
        assert len(result['results']) > 1


def test_filter_intrinsicproperty_formula():
    """Test filter_intrinsicproperty returns a list of CSIDs."""
    qid = cs.filter_intrinsicproperty(formula='C6H6')
    while True:
        status = cs.filter_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_results(qid)
    assert len(results) > 10


def test_filter_intrinsicproperty_mass():
    """Test filter_intrinsicproperty returns a list of CSIDs."""
    qid = cs.filter_intrinsicproperty(monoisotopic_mass=500, monoisotopic_mass_range=0.001)
    while True:
        status = cs.filter_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_results(qid)
    assert len(results) > 10


def test_filter_mass():
    """Test filter_mass returns a list of CSIDs."""
    qid = cs.filter_mass(500, 0.001)
    while True:
        status = cs.filter_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_results(qid)
    assert len(results) > 10


def test_filter_mass_batch():
    """Test filter_mass_batch returns a list of CSIDs."""
    qid = cs.filter_mass_batch(masses=[(12, 0.001), (24, 0.001)])
    while True:
        status = cs.filter_mass_batch_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_mass_batch_results(qid)
    print(results)
    assert len(results) == 2
    for result in results:
        assert 'mass' in result
        assert 'range' in result
        assert 'results' in result
        assert len(result['results']) > 0


def test_filter_smiles():
    """Test filter_smiles returns a list of CSIDs."""
    qid = cs.filter_smiles('c1ccccc1')
    while True:
        status = cs.filter_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    results = cs.filter_results(qid)
    assert len(results) == 1
    assert results[0] == 236  # Benzene ChemSpider ID


def test_filter_sdf():
    """Test filter_results_sdf returns an SDF file."""
    qid = cs.filter_formula('C10H20')
    while True:
        status = cs.filter_status(qid)
        if status['status'] in {'Suspended', 'Failed', 'Not Found', 'Complete'}:
            break
        time.sleep(1)
    sdf = cs.filter_results_sdf(qid)
    assert b'V2000' in sdf
    assert b'$$$$' in sdf


# Tools

def test_convert():
    """Test convert."""
    assert cs.convert('c1ccccc1', 'SMILES', 'InChI') == 'InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H'
    assert cs.convert('InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H', 'InChI', 'InChIKey') == 'UHOVQNZJYSORNB-UHFFFAOYSA-N'
    assert cs.convert('UHOVQNZJYSORNB-UHFFFAOYSA-N', 'InChIKey', 'InChI') == 'InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H'


def test_validate_inchikey():
    """Test validate_inchikey."""
    assert cs.validate_inchikey('UHOVQNZJYSORNB-UHFFFAOYSA-N') is True
    assert cs.validate_inchikey('UHOVQNZJYSORNB-UHFFFAOYSQ-N') is False
    assert cs.validate_inchikey('UHOVQNZJYSORNB-UHFFFAOYSA') is False


# MassSpecAPI

def test_get_databases():
    """Test get_databases returns the list of ChemSpider data sources."""
    with pytest.deprecated_call():
        dbs = cs.get_databases()
        assert all(source in dbs for source in ['Wikipedia', 'ZINC', 'PubChem'])


def test_get_extended_compound_info():
    """Test get_extended_compound_info returns info for a CSID."""
    with pytest.deprecated_call():
        info = cs.get_extended_compound_info(6543)
        assert all(field in info for field in [
            'id', 'smiles', 'formula', 'averageMass', 'molecularWeight', 'monoisotopicMass', 'nominalMass',
            'commonName', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount', 'mol2D'
        ])
        assert all(isinstance(info[field], float) for field in [
            'averageMass', 'molecularWeight', 'monoisotopicMass'
        ])
        assert isinstance(info['id'], int)
        assert all(isinstance(info[field], six.text_type) for field in [
            'smiles', 'formula', 'commonName', 'mol2D'
        ])


def test_get_extended_compound_info_list():
    """Test get_extended_compound_info_list returns info for a list of CSIDs."""
    with pytest.deprecated_call():
        info = cs.get_extended_compound_info_list([6543, 1235, 6084])
        assert len(info) == 3
        assert all(field in info[0] for field in [
            'id', 'smiles', 'formula', 'averageMass', 'molecularWeight', 'monoisotopicMass', 'nominalMass',
            'commonName', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount', 'mol2D'
        ])
        assert all(isinstance(info[0][field], float) for field in [
            'averageMass', 'molecularWeight', 'monoisotopicMass'
        ])
        assert isinstance(info[0]['id'], int)
        assert all(isinstance(info[0][field], six.text_type) for field in [
            'smiles', 'formula', 'commonName', 'mol2D'
        ])


def test_get_extended_mol_compound_info_list():
    """Test get_extended_mol_compound_info_list returns info for a list of CSIDs."""
    with pytest.deprecated_call():
        info = cs.get_extended_mol_compound_info_list([1236], include_external_references=True,
                                                      include_reference_counts=True)
        assert len(info) == 1
        assert all(field in info[0] for field in [
            'id', 'smiles', 'formula', 'averageMass', 'molecularWeight', 'monoisotopicMass', 'nominalMass',
            'commonName', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount', 'mol2D'
        ])
        assert all(isinstance(info[0][field], float) for field in [
            'averageMass', 'molecularWeight', 'monoisotopicMass'
        ])
        assert all(isinstance(info[0][field], int) for field in [
            'id', 'referenceCount', 'dataSourceCount', 'pubMedCount', 'rscCount'
        ])
        assert all(isinstance(info[0][field], six.text_type) for field in [
            'smiles', 'formula', 'commonName', 'mol2D'
        ])


def test_get_record_mol():
    """Test get_record_mol returns a MOL file."""
    with pytest.deprecated_call():
        mol = cs.get_record_mol(6084)
        assert 'V2000' in mol
        assert 'M  END' in mol


# Search

def test_async_simple_search():
    """Test async_simple_search returns a transaction ID."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search('benzene')
        assert re.compile(r'[a-f0-9\-]{20,50}').search(rid)


def test_async_simple_search_ordered():
    """Test async_simple_search returns a transaction ID."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search_ordered('glucose')
        assert re.compile(r'[a-f0-9\-]{20,50}').search(rid)


def test_get_async_search_status():
    """Test get_async_search_status returns the status for a transaction ID."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search('benzene')
        status = cs.get_async_search_status(rid)
        assert status in {
            'Complete', 'Suspended', 'Failed', 'Not Found', 'Unknown', 'Created', 'Scheduled', 'Processing',
            'Suspended', 'PartialResultReady', 'ResultReady'
        }


def test_get_async_search_status_and_count():
    """Test get_async_search_status_and_count returns the status for a transaction ID."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search('benzene')
        while True:
            status = cs.get_async_search_status_and_count(rid)
            if status['status'] in {'Created', 'Scheduled', 'Processing'}:
                continue
            assert status['count'] == 1
            assert status['message'] == 'Found by approved synonym'
            break


def test_get_async_search_result():
    """Test get_async_search_result returns a list of CSIDs."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search('benzene')
        while True:
            status = cs.get_async_search_status(rid)
            if status in {'Created', 'Scheduled', 'Processing'}:
                continue
            assert [c.csid for c in cs.get_async_search_result(rid)] == [236]
            break


def test_get_async_search_result_part():
    """Test get_async_search_result_part returns a list of CSIDs."""
    with pytest.deprecated_call():
        rid = cs.async_simple_search('glucose')
        while True:
            status = cs.get_async_search_status(rid)
            if status in {'Created', 'Scheduled', 'Processing'}:
                continue
            assert len(cs.get_async_search_result_part(rid)) > 6
            assert len(cs.get_async_search_result_part(rid, start=2)) > 2
            assert len(cs.get_async_search_result_part(rid, start=2, count=2)) == 2
            assert len(cs.get_async_search_result_part(rid, start=2, count=99)) > 2
            break


def test_get_compound_info():
    """Test get_compound_info returns info for a CSID."""
    with pytest.deprecated_call():
        info = cs.get_compound_info(123)
        assert all(field in info for field in ['id', 'smiles'])
        assert isinstance(info['id'], int)
        assert isinstance(info['smiles'], six.text_type)


def test_get_compound_thumbnail():
    """Test get_compound_thumbnail returns image data for a CSID."""
    with pytest.deprecated_call():
        img = cs.get_compound_thumbnail(123)
        assert img[:8] == b'\x89PNG\x0d\x0a\x1a\x0a'  # PNG magic number


def test_simple_search():
    """Test simple_search returns a list of CSIDs."""
    assert all(csid in [c.csid for c in cs.simple_search('glucose')] for csid in [5589, 58238, 71358, 96749, 9312824, 9484839])


# Errors


def test_invalid_api_key():
    """Test ChemSpiPyAuthError is raised if a token with invalid format is used."""
    with pytest.raises(errors.ChemSpiPyAuthError):
        mf = ChemSpider('abcde1-1346fa-934a').get_compound(2157).molecular_formula


def test_invalid_api_key2():
    """Test ChemSpiPyAuthError is raised if a fake token with correct format is used."""
    with pytest.raises(errors.ChemSpiPyAuthError):
        mf = ChemSpider('6qBA6lrJycPAYTTcajkkaN02brz5S6Ee').get_compound(2157).molecular_formula


def test_invalid_query_id():
    """Test ChemSpiPyBadRequestError is raised when an invalid query ID is used."""
    with pytest.raises(errors.ChemSpiPyBadRequestError):
        cs.filter_status('xxxxxx')


def test_expired_query_id():
    """Test ChemSpiPyBadRequestError is raised when a valid but expired query ID is used."""
    with pytest.raises(errors.ChemSpiPyBadRequestError):
        cs.filter_status('1a93ee87-acbe-4caa-bc3b-23c3ff39be0f')


def test_fictional_query_id():
    """Test ChemSpiPyBadRequestError is raised when a valid but made up query ID is used."""
    with pytest.raises(errors.ChemSpiPyBadRequestError):
        cs.filter_status('1a93ee87-acbe-4caa-bc3b-23c3ff39be0a')



# -*- coding: utf-8 -*-
"""
chemspipy.search
~~~~~~~~~~~~~~~~

A wrapper for asynchronous search requests.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import datetime
import logging
import threading
import time

from six.moves import range

from . import errors, objects, utils


log = logging.getLogger(__name__)


# TODO: Use Sequence abc metaclass?
class Results(object):
    """Container class to perform a search on a background thread and hold the results when ready."""

    def __init__(self, cs, searchfunc, searchargs, raise_errors=False, max_requests=40):
        """Generally shouldn't be instantiated directly. See :meth:`~chemspipy.api.ChemSpider.search` instead.

        :param ChemSpider cs: ``ChemSpider`` session.
        :param function searchfunc: Search function that returns a transaction ID.
        :param tuple searchargs: Arguments for the search function.
        :param bool raise_errors: If True, raise exceptions. If False, store on ``exception`` property.
        :param int max_requests: Maximum number of times to check if search results are ready.
        """
        log.debug('Results init')
        self._cs = cs
        self._raise_errors = raise_errors
        self._max_requests = max_requests
        self._status = 'Created'
        self._exception = None
        self._qid = None
        self._message = None
        self._start = None
        self._end = None
        self._results = []
        self._searchthread = threading.Thread(name='SearchThread', target=self._search, args=(cs, searchfunc, searchargs))
        self._searchthread.start()

    def _search(self, cs, searchfunc, searchargs):
        """Perform the search and retrieve the results."""
        log.debug('Searching in background thread')
        self._start = datetime.datetime.utcnow()
        try:
            self._qid = searchfunc(*searchargs)
            log.debug('Setting qid: %s' % self._qid)
            for _ in range(self._max_requests):
                log.debug('Checking status: %s' % self._qid)
                status = cs.filter_status(self._qid)
                self._status = status['status']
                self._message = status.get('message', '')
                log.debug(status)
                time.sleep(0.2)
                if status['status'] == 'Complete':
                    break
                elif status['status'] in {'Failed', 'Unknown', 'Suspended', 'Not Found'}:
                    raise errors.ChemSpiPyServerError('Search Failed: %s' % status.get('message', ''))
            else:
                raise errors.ChemSpiPyTimeoutError('Search took too long')
            log.debug('Search success!')
            self._end = datetime.datetime.utcnow()
            if status['count'] > 0:
                self._results = [objects.Compound(cs, csid) for csid in cs.filter_results(self._qid)]
                log.debug('Results: %s', self._results)
            elif not self._message:
                self._message = 'No results found'
        except Exception as e:
            # Catch and store exception so we can raise it in the main thread
            self._exception = e
            self._end = datetime.datetime.utcnow()
            if self._status == 'Created':
                self._status = 'Failed'

    def ready(self):
        """Return True if the search finished.

        :rtype: bool
        """
        return not self._searchthread.is_alive()

    def success(self):
        """Return True if the search finished with no errors.

        :rtype: bool
        """
        return self.ready() and not self._exception

    def wait(self):
        """Block until the search has completed and optionally raise any resulting exception."""
        log.debug('Waiting for search to finish')
        self._searchthread.join()
        if self._exception and self._raise_errors:
            raise self._exception

    @property
    def status(self):
        """Current status string returned by ChemSpider.

        :return: 'Unknown', 'Created', 'Scheduled', 'Processing', 'Suspended', 'PartialResultReady', 'ResultReady'
        :rtype: string
        """
        return self._status

    @property
    def exception(self):
        """Any Exception raised during the search. Blocks until the search is finished."""
        self.wait()  # TODO: If raise_errors=True this will raise the exception when trying to access it?
        return self._exception

    @property
    def qid(self):
        """Search query ID.

        :rtype: string
        """
        return self._qid

    @property
    def message(self):
        """A contextual message about the search. Blocks until the search is finished.

        :rtype: string
        """
        self.wait()
        return self._message

    @property
    def count(self):
        """The number of search results. Blocks until the search is finished.

        :rtype: int
        """
        return len(self)

    @property
    def duration(self):
        """The time taken to perform the search. Blocks until the search is finished.

        :rtype: :py:class:`datetime.timedelta`
        """
        self.wait()
        return self._end - self._start

    @utils.memoized_property
    def sdf(self):
        """Get an SDF containing all the search results.

        :return: SDF containing the search results.
        :rtype: bytes
        """
        self.wait()
        return self._cs.filter_results_sdf(self._qid)

    def __getitem__(self, index):
        """Get a single result or a slice of results. Blocks until the search is finished.

        This means a Results instance can be treated like a normal Python list. For example::

            cs.search('glucose')[2]
            cs.search('glucose')[0:2]

        An IndexError will be raised if the index is greater than the total number of results.
        """
        self.wait()
        return self._results.__getitem__(index)

    def __len__(self):
        self.wait()
        return self._results.__len__()

    def __iter__(self):
        self.wait()
        return iter(self._results)

    def __repr__(self):
        if self.success():
            return 'Results(%s)' % self._results
        else:
            return 'Results(%s)' % self.status



# -*- coding: utf-8 -*-
"""
chemspipy.api
~~~~~~~~~~~~~

Core API for interacting with ChemSpider web services.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import base64
import logging
import sys
import warnings
import zlib

import requests

from . import __version__, errors
from .objects import Compound
from .search import Results


log = logging.getLogger(__name__)


#: Default API URL.
API_URL = 'https://api.rsc.org'
#: Default API version.
API_VERSION = 'v1'

#: 2D coordinate dimensions
MOL2D = '2d'
#: 3D coordinate dimensions
MOL3D = '3d'
#: Both coordinate dimensions
BOTH = 'both'

#: Ascending sort direction
ASCENDING = 'ascending'
#: Descending sort direction
DESCENDING = 'descending'

#: Record ID sort order
RECORD_ID = 'record_id'
#: CSID sort order (same as RECORD_ID, kept for backwards compatibility)
CSID = 'csid'
#: Mass defect sort order
MASS_DEFECT = 'mass_defect'
#: Molecular weight sort order
MOLECULAR_WEIGHT = 'molecular_weight'
#: Reference count sort order
REFERENCE_COUNT = 'reference_count'
#: Datasource count sort order
DATASOURCE_COUNT = 'datasource_count'
#: Pubmed count sort order
PUBMED_COUNT = 'pubmed_count'
#: RSC count sort order
RSC_COUNT = 'rsc_count'


#: Map sort directions to strings required by REST API.
DIRECTIONS = {
    ASCENDING: 'ascending',
    DESCENDING: 'descending'
}

#: Map sort orders to strings required by REST API.
ORDERS = {
    RECORD_ID: 'recordId',
    CSID: 'recordId',
    MASS_DEFECT: 'massDefect',
    MOLECULAR_WEIGHT: 'molecularWeight',
    REFERENCE_COUNT: 'referenceCount',
    DATASOURCE_COUNT: 'dataSourceCount',
    PUBMED_COUNT: 'pubMedCount',
    RSC_COUNT: 'rscCount'
}

#: All available compound details fields.
FIELDS = [
    'SMILES', 'Formula', 'AverageMass', 'MolecularWeight', 'MonoisotopicMass', 'NominalMass', 'CommonName',
    'ReferenceCount', 'DataSourceCount', 'PubMedCount', 'RSCCount', 'Mol2D', 'Mol3D'
]


class ChemSpider(object):
    """Provides access to the ChemSpider API.

    Usage::

        >>> from chemspipy import ChemSpider
        >>> cs = ChemSpider('<YOUR-API-KEY>')

    """

    def __init__(self, api_key, user_agent=None, api_url=API_URL, api_version=API_VERSION):
        """

        :param string api_key: Your ChemSpider API key.
        :param string user_agent: (Optional) Identify your application to ChemSpider servers.
        :param string api_url: (Optional) API server. Default https://api.rsc.org.
        :param string api_version: (Optional) API version. Default v1.
        """
        log.debug('Initializing ChemSpider')
        self.api_url = api_url
        self.http = requests.session()
        self.http.headers['User-Agent'] = user_agent if user_agent else 'ChemSpiPy/{} Python/{} '.format(
            __version__, sys.version.split()[0]
        )
        self.api_key = api_key
        self.api_version = api_version

    def __repr__(self):
        return 'ChemSpider()'

    def request(self, method, api, namespace, endpoint, params=None, json=None):
        """Make a request to the ChemSpider API.

        :param string method: HTTP method.
        :param string api: Top-level API, e.g. compounds.
        :param string namespace: API namespace, e.g. filter, lookups, records, or tools.
        :param string endpoint: Web service endpoint URL.
        :param dict params: Query parameters to add to the URL.
        :param dict json: JSON data to send in the request body.
        :return: Web Service response JSON.
        :rtype: dict
        """
        # Construct request URL
        url = '{}/{}/{}/{}/{}'.format(self.api_url, api, self.api_version, namespace, endpoint)

        # Set apikey header
        headers = {'apikey': self.api_key}

        log.debug('{} : {} : {} : {}'.format(url, headers, params, json))

        # Make request
        r = self.http.request(method, url, params=params, json=json, headers=headers)

        # Raise exception for HTTP errors
        if not r.ok:
            err = {
                400: errors.ChemSpiPyBadRequestError,
                401: errors.ChemSpiPyAuthError,
                404: errors.ChemSpiPyNotFoundError,
                405: errors.ChemSpiPyMethodError,
                413: errors.ChemSpiPyPayloadError,
                429: errors.ChemSpiPyRateError,
                500: errors.ChemSpiPyServerError,
                503: errors.ChemSpiPyUnavailableError
            }.get(r.status_code, errors.ChemSpiPyHTTPError)
            raise err(message=r.reason, http_code=r.status_code)

        log.debug('Request duration: {}'.format(r.elapsed))
        return r.json()

    def get(self, api, namespace, endpoint, params=None):
        """Convenience method for making GET requests.

        :param string api: Top-level API, e.g. compounds.
        :param string namespace: API namespace, e.g. filter, lookups, records, or tools.
        :param string endpoint: Web service endpoint URL.
        :param dict params: Query parameters to add to the URL.
        :return: Web Service response JSON.
        :rtype: dict
        """
        return self.request('GET', api=api, namespace=namespace, endpoint=endpoint, params=params)

    def post(self, api, namespace, endpoint, json=None):
        """Convenience method for making POST requests.

        :param string api: Top-level API, e.g. compounds.
        :param string namespace: API namespace, e.g. filter, lookups, records, or tools.
        :param string endpoint: Web service endpoint URL.
        :param dict json: JSON data to send in the request body.
        :return: Web Service response content.
        :rtype: dict or string
        """
        return self.request('POST', api=api, namespace=namespace, endpoint=endpoint, json=json)

    def get_compound(self, csid):
        """Return a Compound object for a given ChemSpider ID.

        :param string|int csid: ChemSpider ID.
        :return: The Compound with the specified ChemSpider ID.
        :rtype: :class:`~chemspipy.objects.Compound`
        """
        return Compound(self, csid)

    def get_compounds(self, csids):
        """Return a list of Compound objects, given a list ChemSpider IDs.

        :param list[string|int] csids: List of ChemSpider IDs.
        :return: List of Compounds with the specified ChemSpider IDs.
        :rtype: list[:class:`~chemspipy.objects.Compound`]
        """
        return [Compound(self, csid) for csid in csids]

    def search(self, query, order=None, direction=ASCENDING, raise_errors=False):
        """Search ChemSpider for the specified query and return the results.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param string|int query: Search query.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :param bool raise_errors: (Optional) If True, raise exceptions. If False, store on Results ``exception``
                                  property.
        :return: Search Results list.
        :rtype: :class:`~chemspipy.search.Results`
        """
        return Results(self, self.filter_name, (query, order, direction), raise_errors=raise_errors)

    def get_datasources(self):
        """Get the list of datasources in ChemSpider.

        Many other endpoints let you restrict which sources are used to lookup the requested query. Restricting the
        sources makes queries faster.

        :return: List of datasources.
        :rtype: list[string]
        """
        response = self.get(api='compounds', namespace='lookups', endpoint='datasources')
        return response['dataSources']

    def get_details(self, record_id, fields=FIELDS):
        """Get details for a compound record.

        The available fields are listed in :data:`~chemspipy.api.FIELDS`.

        :param int record_id: Record ID.
        :param list[string] fields: (Optional) List of fields to include in the result.
        :return: Record details.
        :rtype: dict
        """
        params = {'fields': ','.join(fields)}
        endpoint = '{}/details'.format(record_id)
        response = self.get(api='compounds', namespace='records', endpoint=endpoint, params=params)
        return response

    def get_details_batch(self, record_ids, fields=FIELDS):
        """Get details for a list of compound records.

        The available fields are listed in :data:`~chemspipy.api.FIELDS`.

        :param list[int] record_ids: List of record IDs (up to 100).
        :param list[string] fields: (Optional) List of fields to include in the results.
        :return: List of record details.
        :rtype: list[dict]
        """
        json = {'recordIds': record_ids, 'fields': fields}
        response = self.post(api='compounds', namespace='records', endpoint='batch', json=json)
        return response['records']

    def get_external_references(self, record_id, datasources=None):
        """Get external references for a compound record.

        Optionally filter the results by data source. Use :meth:`~chemspipy.api.ChemSpider.get_datasources` to get the
        available datasources.

        :param int record_id: Record ID.
        :param list[string] datasources: (Optional) List of datasources to restrict the results to.
        :return: External references.
        :rtype: list[dict]
        """
        params = {}
        if datasources is not None:
            params['dataSources'] = ','.join(datasources)
        endpoint = '{}/externalreferences'.format(record_id)
        response = self.get(api='compounds', namespace='records', endpoint=endpoint, params=params)
        return response['externalReferences']

    def get_image(self, record_id):
        """Get image for a compound record.

        :param int record_id: Record ID.
        :return: Image.
        :rtype: bytes
        """
        endpoint = '{}/image'.format(record_id)
        response = self.get(api='compounds', namespace='records', endpoint=endpoint)
        return base64.b64decode(response['image'])

    def get_mol(self, record_id):
        """Get MOLfile for a compound record.

        :param int record_id: Record ID.
        :return: MOLfile.
        :rtype: string
        """
        endpoint = '{}/mol'.format(record_id)
        response = self.get(api='compounds', namespace='records', endpoint=endpoint)
        return response['sdf']

    def filter_element(self, include_elements, exclude_elements=None, include_all=False, complexity=None, isotopic=None,
                       order=None, direction=None):
        """Search compounds by element.

        Set include_all to true to only consider records that contain all of the elements in ``include_elements``,
        otherwise all records that contain any of the elements will be returned.

        A compound with a complexity of 'multiple' has more than one disconnected system in it or a metal atom or ion.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param list[string] include_elements: List of up to 15 elements to search for compounds containing.
        :param list[string] exclude_elements: List of up to 100 elements to exclude compounds containing.
        :param bool include_all: (Optional) Whether to only include compounds that have all include_elements.
        :param string complexity: (Optional) 'any', 'single', or 'multiple'
        :param string isotopic: (Optional) 'any', 'labeled', or 'unlabeled'.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {
            'includeElements': include_elements,
            'excludeElements': exclude_elements,
            'options': {'includeAll': include_all, 'complexity': complexity, 'isotopic': isotopic},
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        response = self.post(api='compounds', namespace='filter', endpoint='element', json=json)
        return response['queryId']

    def filter_formula(self, formula, datasources=None, order=None, direction=None):
        """Search compounds by formula.

        Optionally filter the results by data source. Use :meth:`~chemspipy.api.ChemSpider.get_datasources` to get the
        available datasources.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param string formula: Molecular formula.
        :param list[string] datasources: (Optional) List of datasources to restrict the results to.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {
            'formula': formula,
            'dataSources': datasources,
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        response = self.post(api='compounds', namespace='filter', endpoint='formula', json=json)
        return response['queryId']

    def filter_formula_batch(self, formulas, datasources=None, order=None, direction=None):
        """Search compounds with a list of formulas.

        Optionally filter the results by data source. Use :meth:`~chemspipy.api.ChemSpider.get_datasources` to get the
        available datasources.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param list[string] formulas: Molecular formula.
        :param list[string] datasources: (Optional) List of datasources to restrict the results to.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_formula_batch_status`` and ``filter_formula_batch_results``.
        :rtype: string
        """
        json = {
            'formulas': formulas,
            'dataSources': datasources,
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        response = self.post(api='compounds', namespace='filter', endpoint='formula/batch', json=json)
        return response['queryId']

    def filter_formula_batch_status(self, query_id):
        """Get formula batch filter status using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous formula batch filter request.
        :return: Status dict with 'status', 'count', and 'message' fields.
        :rtype: dict
        """
        endpoint = 'formula/batch/{}/status'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return response

    def filter_formula_batch_results(self, query_id):
        """Get formula batch filter results using a query ID that was returned by a previous filter request.

        Each result is a dict containing a ``formula`` key and a ``results`` key.

        :param string query_id: Query ID from a previous formula batch filter request.
        :return: List of results.
        :rtype: list[dict]
        """
        endpoint = 'formula/batch/{}/results'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return response['batchResults']

    def filter_inchi(self, inchi):
        """Search compounds by InChI.

        :param string inchi: InChI.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {'inchi': inchi}
        response = self.post(api='compounds', namespace='filter', endpoint='inchi', json=json)
        return response['queryId']

    def filter_inchikey(self, inchikey):
        """Search compounds by InChIKey.

        :param string inchikey: InChIKey.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {'inchikey': inchikey}
        response = self.post(api='compounds', namespace='filter', endpoint='inchikey', json=json)
        return response['queryId']

    def filter_intrinsicproperty(self, formula=None, molecular_weight=None, nominal_mass=None, average_mass=None,
                                 monoisotopic_mass=None, molecular_weight_range=None, nominal_mass_range=None,
                                 average_mass_range=None, monoisotopic_mass_range=None, complexity=None, isotopic=None,
                                 order=None, direction=None):
        """Search compounds by intrinsic property, such as formula and mass.

        At least one of formula, molecular_weight, nominal_mass, average_mass, monoisotopic_mass must be specified.

        A compound with a complexity of 'multiple' has more than one disconnected system in it or a metal atom or ion.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param string formula: Molecular formula.
        :param float molecular_weight: Molecular weight.
        :param float nominal_mass: Nominal mass.
        :param float average_mass: Average mass.
        :param float monoisotopic_mass: Monoisotopic mass.
        :param float molecular_weight_range: Molecular weight range.
        :param float nominal_mass_range: Nominal mass range.
        :param float average_mass_range: Average mass range.
        :param float monoisotopic_mass_range: Monoisotopic mass range.
        :param string complexity: (Optional) 'any', 'single', or 'multiple'
        :param string isotopic: (Optional) 'any', 'labeled', or 'unlabeled'.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {
            'formula': formula,
            'options': {'complexity': complexity, 'isotopic': isotopic},
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        if molecular_weight is not None and molecular_weight_range is not None:
            json['molecularWeight'] = {'mass': molecular_weight, 'range': molecular_weight_range}
        if nominal_mass is not None and nominal_mass_range is not None:
            json['nominalMass'] = {'mass': nominal_mass, 'range': nominal_mass_range}
        if average_mass is not None and average_mass_range is not None:
            json['averageMass'] = {'mass': average_mass, 'range': average_mass_range}
        if monoisotopic_mass is not None and monoisotopic_mass_range is not None:
            json['monoisotopicMass'] = {'mass': monoisotopic_mass, 'range': monoisotopic_mass_range}
        response = self.post(api='compounds', namespace='filter', endpoint='intrinsicproperty', json=json)
        return response['queryId']

    def filter_mass(self, mass, mass_range, datasources=None, order=None, direction=None):
        """Search compounds by mass.

        Filter to compounds within ``mass_range`` of the given ``mass``.

        Optionally filter the results by data source. Use :meth:`~chemspipy.api.ChemSpider.get_datasources` to get the
        available datasources.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param float mass: Mass between 1 and 11000 Atomic Mass Units.
        :param float mass_range: Mass range between 0.0001 and 100 Atomic Mass Units.
        :param list[string] datasources: (Optional) List of datasources to restrict the results to.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {
            'mass': mass,
            'range': mass_range,
            'dataSources': datasources,
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        response = self.post(api='compounds', namespace='filter', endpoint='mass', json=json)
        return response['queryId']

    def filter_mass_batch(self, masses, datasources=None, order=None, direction=None):
        """Search compounds with a list of masses and mass ranges.

        The ``masses`` parameter should be a list of tuples, each with two elements: A mass, and a mass range::

            qid = cs.filter_mass_batch(masses=[(12, 0.001), (24, 0.001)])

        Optionally filter the results by data source. Use :meth:`~chemspipy.api.ChemSpider.get_datasources` to get the
        available datasources.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param list[tuple[float, float]] masses: List of (mass, range) tuples.
        :param list[string] datasources: (Optional) List of datasources to restrict the results to.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_formula_batch_status`` and ``filter_formula_batch_results``.
        :rtype: string
        """
        masses = [{'mass': m, 'range': r} for m, r in masses]
        json = {
            'masses': masses,
            'dataSources': datasources,
            'orderBy': ORDERS.get(order),
            'orderDirection': DIRECTIONS.get(direction)
        }
        response = self.post(api='compounds', namespace='filter', endpoint='mass/batch', json=json)
        return response['queryId']

    def filter_mass_batch_status(self, query_id):
        """Get formula batch filter status using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous formula batch filter request.
        :return: Status dict with 'status', 'count', and 'message' fields.
        :rtype: dict
        """
        endpoint = 'mass/batch/{}/status'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return response

    def filter_mass_batch_results(self, query_id):
        """Get formula batch filter results using a query ID that was returned by a previous filter request.

        Each result is a dict containing a ``formula`` key and a ``results`` key.

        :param string query_id: Query ID from a previous formula batch filter request.
        :return: List of results.
        :rtype: list[dict]
        """
        endpoint = 'mass/batch/{}/results'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return response['batchResults']

    def filter_name(self, name, order=None, direction=None):
        """Search compounds by name.

        The accepted values for ``order`` are: :data:`~chemspipy.api.RECORD_ID`, :data:`~chemspipy.api.MASS_DEFECT`,
        :data:`~chemspipy.api.MOLECULAR_WEIGHT`, :data:`~chemspipy.api.REFERENCE_COUNT`,
        :data:`~chemspipy.api.DATASOURCE_COUNT`, :data:`~chemspipy.api.PUBMED_COUNT` or
        :data:`~chemspipy.api.RSC_COUNT`.

        :param string name: Compound name.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {'name': name, 'orderBy': ORDERS.get(order),  'orderDirection': DIRECTIONS.get(direction)}
        response = self.post(api='compounds', namespace='filter', endpoint='name', json=json)
        return response['queryId']

    def filter_smiles(self, smiles):
        """Search compounds by SMILES.

        :param string smiles: Compound SMILES.
        :return: Query ID that may be passed to ``filter_status`` and ``filter_results``.
        :rtype: string
        """
        json = {'smiles': smiles}
        response = self.post(api='compounds', namespace='filter', endpoint='smiles', json=json)
        return response['queryId']

    def filter_status(self, query_id):
        """Get filter status using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous filter request.
        :return: Status dict with 'status', 'count', and 'message' fields.
        :rtype: dict
        """
        endpoint = '{}/status'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return response

    def filter_results(self, query_id, start=None, count=None):
        """Get filter results using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous filter request.
        :param int start: Zero-based results offset.
        :param int count: Number of results to return.
        :return: List of results.
        :rtype: list[int]
        """
        endpoint = '{}/results'.format(query_id)
        params = {'start': start, 'count': count}
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint, params=params)
        return response['results']

    def filter_results_sdf(self, query_id):
        """Get filter results as SDF file using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous filter request.
        :return: SDF file containing the results.
        :rtype: bytes
        """
        endpoint = '{}/results/sdf'.format(query_id)
        response = self.get(api='compounds', namespace='filter', endpoint=endpoint)
        return zlib.decompress(base64.b64decode(response['results']), 16 + zlib.MAX_WBITS)

    def convert(self, input, input_format, output_format):
        """Convert a chemical from one format to another.

        Format: ``SMILES``, ``InChI``, ``InChIKey`` or ``Mol``.

        Allowed conversions: from InChI to InChIKey, from InChI to Mol file, from InChI to SMILES, from InChIKey to
        InChI, from InChIKey to Mol file, from Mol file to InChI, from Mol file to InChIKey, from SMILES to InChI.

        :param string input: Input chemical.
        :param string input_format: Input format.
        :param string output_format: Output format.
        :return: Input chemical in output format.
        :rtype: string
        """
        json = {'input': input, 'inputFormat': input_format, 'outputFormat': output_format}
        response = self.post(api='compounds', namespace='tools', endpoint='convert', json=json)
        return response['output']

    def validate_inchikey(self, inchikey):
        """Return whether ``inchikey`` is valid.

        :param string inchikey: The InChIKey to validate.
        :return: Whether the InChIKey is valid.
        :rtype: bool
        """
        json = {'inchikey': inchikey}
        try:
            response = self.post(api='compounds', namespace='tools', endpoint='validate/inchikey', json=json)
            return response['valid']
        except errors.ChemSpiPyHTTPError:
            return False

    def get_databases(self):
        """Get the list of datasources in ChemSpider.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.get_datasources` instead.

        """
        warnings.warn('Use get_datasources instead of get_databases.', DeprecationWarning)
        return self.get_datasources()

    def get_extended_compound_info(self, csid):
        """Get extended record details for a CSID.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.get_details` instead.

        :param string|int csid: ChemSpider ID.
        """
        warnings.warn('Use get_details instead of get_extended_compound_info.', DeprecationWarning)
        return self.get_details(record_id=csid)

    def get_extended_compound_info_list(self, csids):
        """Get extended record details for a list of CSIDs.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.get_details_batch` instead.

        :param list[string|int] csids: ChemSpider IDs.
        """
        warnings.warn('Use get_details_batch instead of get_extended_compound_info.', DeprecationWarning)
        return self.get_details_batch(record_ids=csids)

    def get_extended_mol_compound_info_list(self, csids, mol_type=MOL2D, include_reference_counts=False,
                                            include_external_references=False):
        """Get extended record details (including MOL) for a list of CSIDs.

        A maximum of 250 CSIDs can be fetched per request.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.get_details_batch` instead.

        :param list[string|int] csids: ChemSpider IDs.
        :param string mol_type: :data:`~chemspipy.api.MOL2D`, :data:`~chemspipy.api.MOL3D` or
                                :data:`~chemspipy.api.BOTH`.
        :param bool include_reference_counts: Whether to include reference counts.
        :param bool include_external_references: Whether to include external references.
        """
        warnings.warn('Use get_details_batch instead of get_extended_mol_compound_info_list.', DeprecationWarning)
        return self.get_details_batch(record_ids=csids)

    def get_record_mol(self, csid, calc3d=False):
        """Get ChemSpider record in MOL format.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.get_mol` instead.

        :param string|int csid: ChemSpider ID.
        :param bool calc3d: Whether 3D coordinates should be calculated before returning record data.
        """
        warnings.warn('Use get_mol instead of get_record_mol.', DeprecationWarning)
        if calc3d:
            warnings.warn('calc3d parameter for get_record_mol is no longer supported.', DeprecationWarning)
        return self.get_mol(record_id=csid)

    def async_simple_search(self, query):
        """Search ChemSpider with arbitrary query, returning results in order of the best match found.

        This method returns a transaction ID which can be used with other methods to get search status and results.

        .. deprecated:: 2.0.0
           Use :py:meth:`~chemspipy.api.ChemSpider.filter_name` instead.

        :param string query: Search query - a name, SMILES, InChI, InChIKey, CSID, etc.
        :return: Transaction ID.
        :rtype: string
        """
        warnings.warn('Use filter_name instead of async_simple_search.', DeprecationWarning)
        return self.filter_name(name=query)

    def async_simple_search_ordered(self, query, order=CSID, direction=ASCENDING):
        """Search ChemSpider with arbitrary query, returning results with a custom order.

        This method returns a transaction ID which can be used with other methods to get search status and results.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.filter_name` instead.

        :param string query: Search query - a name, SMILES, InChI, InChIKey, CSID, etc.
        :param string order: (Optional) Field to sort the result by.
        :param string direction: (Optional) :data:`~chemspipy.api.ASCENDING` or :data:`~chemspipy.api.DESCENDING`.
        :return: Transaction ID.
        :rtype: string
        """
        warnings.warn('Use filter_name instead of async_simple_search.', DeprecationWarning)
        return self.filter_name(name=query, order=order, direction=direction)

    def get_async_search_status(self, rid):
        """Check the status of an asynchronous search operation.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.filter_status` instead.

        :param string rid: A transaction ID, returned by an asynchronous search method.
        :return: Unknown, Created, Scheduled, Processing, Suspended, PartialResultReady, ResultReady, Failed,
                  TooManyRecords
        :rtype: string
        """
        warnings.warn('Use filter_status instead of get_async_search_status.', DeprecationWarning)
        return self.filter_status(query_id=rid)['status']

    def get_async_search_status_and_count(self, rid):
        """Check the status of an asynchronous search operation. If ready, a count and message are also returned.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.filter_status` instead.

        :param string rid: A transaction ID, returned by an asynchronous search method.
        :rtype: dict
        """
        warnings.warn('Use filter_status instead of get_async_search_status_and_count.', DeprecationWarning)
        return self.filter_status(query_id=rid)

    def get_async_search_result(self, rid):
        """Get the results from a asynchronous search operation.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.filter_results` instead.

        :param string rid: A transaction ID, returned by an asynchronous search method.
        :return: A list of Compounds.
        :rtype: list[:class:`~chemspipy.objects.Compound`]
        """
        warnings.warn('Use filter_results instead of get_async_search_result.', DeprecationWarning)
        results = self.filter_results(query_id=rid)
        return [Compound(self, record_id) for record_id in results]

    def get_async_search_result_part(self, rid, start=0, count=-1):
        """Get a slice of the results from a asynchronous search operation.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.filter_results` instead.

        :param string rid: A transaction ID, returned by an asynchronous search method.
        :param int start: The number of results to skip.
        :param int count: The number of results to return. -1 returns all through to end.
        :return: A list of Compounds.
        :rtype: list[:class:`~chemspipy.objects.Compound`]
        """
        warnings.warn('Use filter_results instead of get_async_search_result_part.', DeprecationWarning)
        if count == -1:
            count = None
        results = self.filter_results(query_id=rid, start=start, count=count)
        return [Compound(self, record_id) for record_id in results]

    def get_compound_info(self, csid):
        """Get SMILES, StdInChI and StdInChIKey for a given CSID.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.get_details` instead.

        :param string|int csid: ChemSpider ID.
        :rtype: dict
        """
        warnings.warn('Use get_details instead of get_compound_info.', DeprecationWarning)
        return self.get_details(record_id=csid)

    def get_compound_thumbnail(self, csid):
        """Get PNG image as binary data.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.get_image` instead.

        :param string|int csid: ChemSpider ID.
        :rtype: bytes
        """
        warnings.warn('Use get_image instead of get_compound_thumbnail.', DeprecationWarning)
        return self.get_image(record_id=csid)

    def simple_search(self, query):
        """Search ChemSpider with arbitrary query.

        .. deprecated:: 2.0.0
           Use :meth:`~chemspipy.api.ChemSpider.search` instead.

        :param string query: Search query - a chemical name.
        :return: Search Results list.
        :rtype: :class:`~chemspipy.search.Results`
        """
        warnings.warn('Use search instead of simple_search.', DeprecationWarning)
        return self.search(query=query)



# -*- coding: utf-8 -*-
"""
chemspipy.objects
~~~~~~~~~~~~~~~~~

Objects returned by ChemSpiPy API methods.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import warnings


from .utils import memoized_property


class Compound(object):
    """ A class for retrieving and caching details about a specific ChemSpider record.

    The purpose of this class is to provide access to various parts of the ChemSpider API that return information about
    a compound given its ChemSpider ID. Information is loaded lazily when requested, and cached for future access.
    """

    def __init__(self, cs, record_id):
        """

        :param ChemSpider cs: ``ChemSpider`` session.
        :param int|string record_id: Compound record ID.
        """
        self._cs = cs
        self._record_id = int(record_id)
        # TODO: Allow optional initialize  with a record-type response from the API (kwarg or class method from_dict?).

    def __eq__(self, other):
        return isinstance(other, Compound) and self.csid == other.csid

    def __repr__(self):
        return 'Compound(%r)' % self.csid

    def _repr_png_(self):
        """For IPython notebook, display 2D image."""
        return self.image

    @property
    def record_id(self):
        """Compound record ID.

        :rtype: int
        """
        return self._record_id

    @property
    def csid(self):
        """ChemSpider ID.

        .. deprecated:: 2.0.0
           Use :py:attr:`~chemspipy.objects.Compound.record_id` instead.

        :rtype: int
        """
        warnings.warn('Use record_id instead of csid.', DeprecationWarning)
        return self._record_id

    @property
    def image_url(self):
        """Return the URL of a PNG image of the 2D chemical structure.

        :rtype: string
        """
        return 'http://www.chemspider.com/ImagesHandler.ashx?id=%s' % self.record_id

    @memoized_property
    def _details(self):
        """Request compound info and cache the result."""
        return self._cs.get_details(self.record_id)

    @property
    def molecular_formula(self):
        """Return the molecular formula for this Compound.

        :rtype: string
        """
        return self._details['formula']

    @property
    def smiles(self):
        """Return the SMILES for this Compound.

        :rtype: string
        """
        return self._details['smiles']

    # TODO: Convert tool to get inchi?

    @property
    def stdinchi(self):
        """Return the Standard InChI for this Compound.

        .. deprecated:: 2.0.0
           Use :py:attr:`~chemspipy.objects.Compound.inchi` instead.

        :rtype: string
        """
        warnings.warn('Use inchi instead of stdinchi.', DeprecationWarning)
        return self.inchi

    @property
    def stdinchikey(self):
        """Return the Standard InChIKey for this Compound.

        .. deprecated:: 2.0.0
           Use :py:attr:`~chemspipy.objects.Compound.inchikey` instead.

        :rtype: string
        """
        warnings.warn('Use inchikey instead of stdinchikey.', DeprecationWarning)
        return self.inchikey

    @property
    def inchi(self):
        """Return the InChI for this Compound.

        :rtype: string
        """
        return self._cs.convert(self.mol_2d, 'Mol', 'InChI')

    @property
    def inchikey(self):
        """Return the InChIKey for this Compound.

        :rtype: string
        """
        return self._cs.convert(self.mol_2d, 'Mol', 'InChIKey')

    @property
    def average_mass(self):
        """Return the average mass of this Compound.

        :rtype: float
        """
        return self._details['averageMass']

    @property
    def molecular_weight(self):
        """Return the molecular weight of this Compound.

        :rtype: float
        """
        return self._details['molecularWeight']

    @property
    def monoisotopic_mass(self):
        """Return the monoisotopic mass of this Compound.

        :rtype: float
        """
        return self._details['monoisotopicMass']

    @property
    def nominal_mass(self):
        """Return the nominal mass of this Compound.

        :rtype: float
        """
        return self._details['nominalMass']

    @property
    def common_name(self):
        """Return the common name for this Compound.

        :rtype: string
        """
        return self._details['commonName']

    @memoized_property
    def mol_2d(self):
        """Return the MOL file for this Compound with 2D coordinates.

        :rtype: string
        """
        return self._details['mol2D']

    @memoized_property
    def mol_3d(self):
        """Return the MOL file for this Compound with 3D coordinates.

        :rtype: string
        """
        return self._details['mol3D']

    @memoized_property
    def image(self):
        """Return a 2D depiction of this Compound.

        :rtype: bytes
        """
        return self._cs.get_image(self.record_id)

    @memoized_property
    def external_references(self):
        """Return external references for this Compound.

        :rtype: list[dict]
        """
        return self._cs.get_external_references(self.record_id)



# -*- coding: utf-8 -*-
"""
ChemSpiPy
~~~~~~~~~

Python wrapper for the ChemSpider API.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division


__author__ = 'Matt Swain'
__email__ = 'm.swain@me.com'
__version__ = '2.0.0'
__license__ = 'MIT'


from .api import ChemSpider, MOL2D, MOL3D, BOTH, ASCENDING, DESCENDING, RECORD_ID, CSID, MASS_DEFECT, MOLECULAR_WEIGHT
from .api import REFERENCE_COUNT, DATASOURCE_COUNT, PUBMED_COUNT, RSC_COUNT, FIELDS
from .objects import Compound
from .search import Results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup


if os.path.exists('README.rst'):
    long_description = open('README.rst').read()
else:
    long_description = '''A simple Python wrapper around the ChemSpider Web Services.'''

setup(
    name='ChemSpiPy',
    version='2.0.0',
    author='Matt Swain',
    author_email='m.swain@me.com',
    license='MIT',
    url='https://github.com/mcs07/ChemSpiPy',
    packages=['chemspipy'],
    description='A simple Python wrapper around the ChemSpider Web Services.',
    long_description=long_description,
    keywords='chemistry cheminformatics chemspider rsc rest api',
    zip_safe=False,
    install_requires=['requests', 'six'],
    tests_require=['pytest'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
"""
test_search
~~~~~~~~~~~

Test the search wrapper.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import logging
import os

import pytest

from chemspipy import ChemSpider, errors, ASCENDING, DESCENDING, CSID, REFERENCE_COUNT, MOLECULAR_WEIGHT


logging.basicConfig(level=logging.WARN, format='%(levelname)s:%(name)s:(%(threadName)-10s):%(message)s')
logging.getLogger('chemspipy').setLevel(logging.DEBUG)

# API key is retrieved from environment variables
CHEMSPIDER_API_KEY = os.environ['CHEMSPIDER_API_KEY']
cs = ChemSpider(CHEMSPIDER_API_KEY)


def test_search_smiles():
    """Test SMILES input to search."""
    results = cs.search('O=C(OCC)C')
    assert results.ready() is False
    results.wait()
    assert results.ready() is True
    assert results.success() is True
    assert results.message == 'Found by conversion query string to chemical structure (full match)'
    assert results[0].csid == 8525
    assert results.duration.total_seconds() > 0


def test_search_csid():
    """Test ChemSpider ID input to search."""
    results = cs.search(8525)
    assert results.message == 'Found by CSID'
    assert len(results) == 1
    assert repr(results) == 'Results([Compound(8525)])'
    assert results[0].csid == 8525


def test_search_name():
    """Test name input to search."""
    results = cs.search('propanol')
    assert results.message == 'Found by approved synonym'
    assert results[0].csid == 1004


def test_search_iter():
    """Test iteration of search results."""
    for result in cs.search('glucose'):
        assert isinstance(result.csid, int)


def test_search_ordered_csid():
    """Test search results ordered by CSID."""
    results = cs.search('glucose', order=CSID)
    assert list(results) == sorted(results, key=lambda x: x.csid)


def test_search_ordered_csid_descending():
    """Test search results ordered by CSID and direction descending."""
    results = cs.search('glucose', order=CSID, direction=DESCENDING)
    assert list(results) == sorted(results, key=lambda x: x.csid, reverse=True)


def test_search_ordered_ref_descending():
    """Test search results ordered by CSID and direction descending."""
    results = cs.search('glucose', order=REFERENCE_COUNT, direction=DESCENDING)
    assert [result.csid for result in results]


def test_search_ordered_weight_ascending():
    """Test search results ordered by CSID and direction descending."""
    results = cs.search('P', order=MOLECULAR_WEIGHT, direction=ASCENDING)
    assert list(results) == sorted(results, key=lambda x: x.molecular_weight)


def test_search_no_results():
    """Test name input to search."""
    results = cs.search('aergherguyaelrgiaubrfawyef')
    assert results.message == 'No results found'
    assert results.ready() is True
    assert results.success() is True
    assert len(results) == 0


def test_too_high_index():
    """Test IndexError is raised for a too high index."""
    with pytest.raises(IndexError):
        result = cs.search('glucose')[7843]


def test_search_failed():
    """Test ChemSpiPyServerError is raised for an invalid SMILES."""
    results = cs.search('O=C(OCC)C*')
    results.wait()
    assert isinstance(results.exception, errors.ChemSpiPyBadRequestError)
    assert results.status == 'Failed'
    assert repr(results) == 'Results(Failed)'
    assert results.ready() is True
    assert results.success() is False
    assert results.count == 0
    assert results.duration.total_seconds() > 0


def test_search_exception():
    """Test ChemSpiPyServerError is raised for an invalid SMILES."""
    with pytest.raises(errors.ChemSpiPyBadRequestError):
        results = cs.search('O=C(OCC)C*', raise_errors=True)
        results.wait()



# -*- coding: utf-8 -*-
"""
chemspipy.errors
~~~~~~~~~~~~~~~~

Exceptions raised by ChemSpiPy.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division


class ChemSpiPyError(Exception):
    """Root ChemSpiPy Exception."""
    pass


class ChemSpiPyHTTPError(ChemSpiPyError):
    """Base exception to handle HTTP errors."""

    #: Default message if none supplied. Override in subclasses.
    MESSAGE = 'ChemSpiPy Error'
    HTTP_CODE = None

    def __init__(self, message=None, http_code=None, *args, **kwargs):
        """

        :param string|bytes message: Error message.
        :param http_code: HTTP code.
        """

        # Decode message to unicode if necessary
        if isinstance(message, bytes):
            try:
                message = message.decode('utf-8')
            except UnicodeDecodeError:
                message = message.decode('iso-8859-1')

        self.message = message if message is not None else self.MESSAGE
        self.http_code = http_code if http_code is not None else self.HTTP_CODE
        super(ChemSpiPyHTTPError, self).__init__(*args, **kwargs)

    def __repr__(self):
        args = 'message={!r}'.format(self.message)
        if self.http_code is not None:
            args += ', http_code={!r}'.format(self.http_code)
        return '{}({})'.format(self.__class__.__name__, args)

    def __str__(self):
        return self.message


class ChemSpiPyBadRequestError(ChemSpiPyHTTPError):
    """Raised for a bad request."""
    MESSAGE = 'Bad request.'
    HTTP_CODE = 400


class ChemSpiPyAuthError(ChemSpiPyHTTPError):
    """Raised when API key authorization fails."""
    MESSAGE = 'Unauthorized.'
    HTTP_CODE = 401


class ChemSpiPyNotFoundError(ChemSpiPyHTTPError):
    """Raised when the requested resource was not found."""
    MESSAGE = 'Not found.'
    HTTP_CODE = 404


class ChemSpiPyMethodError(ChemSpiPyHTTPError):
    """Raised when an invalid HTTP method is used."""
    MESSAGE = 'Method Not Allowed.'
    HTTP_CODE = 405


class ChemSpiPyPayloadError(ChemSpiPyHTTPError):
    """Raised when a request payload is too large."""
    MESSAGE = 'Payload Too Large.'
    HTTP_CODE = 413


class ChemSpiPyRateError(ChemSpiPyHTTPError):
    """Raised when too many requests are sent in a given amount of time."""
    MESSAGE = 'Too Many Requests.'
    HTTP_CODE = 429


class ChemSpiPyServerError(ChemSpiPyHTTPError):
    """Raised when an internal server error occurs."""
    MESSAGE = 'Internal Server Error.'
    HTTP_CODE = 500


class ChemSpiPyUnavailableError(ChemSpiPyHTTPError):
    """Raised when the service is temporarily unavailable."""
    MESSAGE = 'Service Unavailable.'
    HTTP_CODE = 503


class ChemSpiPyTimeoutError(ChemSpiPyError):
    """Raised when an asynchronous request times out."""
    pass



# -*- coding: utf-8 -*-
"""
chemspipy.utils
~~~~~~~~~~~~~~~

Miscellaneous utility functions.

"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
import datetime
import functools


def memoized_property(fget):
    """Decorator to create memoized properties."""
    attr_name = '_{}'.format(fget.__name__)

    @functools.wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)
    return property(fget_memoized)


def timestamp(ts):
    """Create a datetime object from a timestamp string."""
    fmt = '%Y-%m-%dT%H:%M:%S.%f' if '.' in ts else '%Y-%m-%dT%H:%M:%S'
    return datetime.datetime.strptime(ts, fmt)


def duration(ts):
    """Create a timedelta object from a duration string."""
    fmt = '%H:%M:%S.%f' if '.' in ts else '%H:%M:%S'
    dt = datetime.datetime.strptime(ts, fmt)
    return datetime.timedelta(hours=dt.hour, minutes=dt.minute, seconds=dt.second, microseconds=dt.microsecond)
